/*============================================================================//
    SatPlan.cpp: where main() is located.

    Implementation: Shane J. Neph, June 2004, University of Washington
//============================================================================*/

//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.

// Files included
#include "ByLine.h"
#include "Assertion.h"
#include "Conversion.h"
#include "CommonErrors.h"
#include "Input.h"
#include "StandardFiles.h"
#include "StringAlgorithms.h"
#include "Timer.h"
#include "Typify.h"

// unix-specific headers
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>

//Parser and cost optimal planner headers
#include "Arguments.hh"
#include "global.hh"
#include "Problem.hh"
#include "Planner.hh"
#include "Planner_templates.hh"
#include "CostPlanner.hh"
#include "State.hh"
#include "PredicatesAndPropositions.hh"

enum Status { SAT = 0, UNSAT = 2, ERROR = -1 };

namespace {

    // Function Prototypes
    void AddCNFInfo(const Input&);
    void CheckProgramTime(const Input&, const Timer&);
    void MoveSolution(const std::string&);
    void RedirectOutput();
    
    typedef Exc::Exception<1> Unsolved;

    enum ReturnValues 
      { GOOD = 0, NOSOLVE = 1, NOMEMORYTIME = 2, NOGOOD = 3, BADUSAGE = 4 };

    std::string solverResults, finalSolutionFile;
    static const std::string undefined = "undefined";
    std::string exeDir = undefined;

    std::map<long, long> reducibles;
} // unnamed

//NATHAN
extern int yyparse();
extern void initialiseLexer(istream*, ostream* );
extern void deleteLexer();
extern Planning::Problem problem;


//NATHAN
//Create the required objects and parse the domain and problem files.
bool local_parse_function(const string &domainFileName, const string &problemFileName)
{
    ifstream domainFile;
    domainFile.open(domainFileName.c_str(), ifstream::in);
    if(!domainFile.is_open()){
	    cerr << "Couldn't open domain definition.\n";
	    return false;
    }
    initialiseLexer(&domainFile, &cout);
    if(yyparse()){
	    cerr << "Parsing of domain file :: "<<domainFileName<<" failed!\n";
	    return false;
    }
    deleteLexer();
    domainFile.close();
    ifstream problemFile;
    problemFile.open(problemFileName.c_str(), ifstream::in);
    if(!problemFile.is_open()){
	    cerr << "Couldn't open problem definition.\n";
	    return false;
    }
    initialiseLexer(&problemFile, &cout);
    if(yyparse()){
	    cerr << "Parsing of problem file :: "<<problemFileName<<" failed!\n";
	    return false;
    }
    deleteLexer();
    problemFile.close();

    //Types
    problem.domain.configureTypes();
    problem.domain.unwindTypes();
    //Constants
    problem.domain.configureConstants();
    problem.domain.unwindConstants();
    //Constants -- objects
    problem.configureObjects();
    problem.unwindObjects();

    problem.domain.initialiseMirrors();
    problem.initialiseMirrors();

    /*Decide what predicate symbols are fluent, and their type of fluency*/
    problem.domain.processActionsForSymbolTypes();
    
    return true;
} 

//NATHAN
//Generate the CNF Encoding
//return true if it was generated
Status GenEncoding(Input& input_, long planLevel, int prevBestScore) {
    string args;
    string domain = input_.Domain();
    string prob = input_.Problem();
    string cnfName = input_.CNF();
    string varFileName = input_.VarFile();
    
    args = "./bb -o " + domain
        + " -f " + prob + " -b " + cnfName + " -l " + 
        convert<std::string>(planLevel) + " -V " + varFileName +
        " -z " + convert<std::string>(prevBestScore);

    //cout << "Calling bb with: " << args << endl;

    int sysCall = system(args.c_str());

    if ( WIFSIGNALED(sysCall) ) {
        Assert<CE::ProgramError>(false, "BB failed... aborting");
    }

    if ( UNSAT == WEXITSTATUS(sysCall) )
        return(UNSAT);
    else if ( SAT == WEXITSTATUS(sysCall) ) {
        // SAT is sometimes returned even when a resource is exhausted."
        std::ifstream ifile(cnfName.c_str());
        Assert<CE::MemoryLimit>(ifile, "BB failed... aborting");
        return(SAT);
    }
    std::ifstream ifile(cnfName.c_str());
    Assert<CE::MemoryLimit>(ifile, "BB failed... aborting");
    return(ERROR);
}

//NATHAN - modified from the previous SolveCNF() function
//This function attempts to solve a previously generated CNF file with rsat
bool SolveCNF(const Input& input, int &costOfSoln, std::vector<int> &varAssignments)
{
    string exeName = input.Program();
    string cnfName = input.CNF();
    string solverResult = DEFAULT_SAT_RESULT_NAME;
    // Try to solve CNF sentence
    try {       
        stringstream timeout_os;
        timeout_os << input.SolverLim();
        string toRun = "./" + exeName + " " + cnfName + " -s -o -t " + timeout_os.str() + 
                " > " + solverResult;
        // Run solver and check status information
        Timer timer;
        timer.Start();
        int rtn = std::system(toRun.c_str());
        timer.Stop();
        
        if ( WIFSIGNALED(rtn) ) {
            int tmpRtn = WTERMSIG(rtn);
            Assert<CE::UserAbort>(tmpRtn != SIGQUIT, "user abort");
            Assert<CE::UserAbort>(tmpRtn != SIGINT, "user abort");
            std::string err = "memory or time limit exceeded";
            Assert<CE::TimeOut>(tmpRtn != SIGXCPU, "timeout");
            Assert<CE::MemoryLimit>(tmpRtn != SIGXFSZ, "memory exhausted");
            Assert<CE::MemoryLimit>(tmpRtn != SIGSEGV, "memory exhausted");
            err = convert<std::string>(tmpRtn);
            err = "Unchecked Signal Generated (# " + err + ") ...aborting";
            Assert<CE::AbortedMission>(false, err);
        }
        // Let solver evaluate return value
        bool isOK = true;
        ifstream ifile(solverResult.c_str());
        if ( !ifile ) return(false);
        
        //Read the solution
        int rsResult = -1;
        ifile >> rsResult;
        
        //cout << "rsResult...: " << rsResult << endl;
        
        costOfSoln = rsResult;
        if (rsResult >= 0) {
            //extract the actual details of the solution from the ifile
            string tStr;
            ifile >> tStr;
            assert(tStr == "v");
            while (ifile >> rsResult) {
               if (rsResult == 0) break;
               varAssignments.push_back(rsResult);
            }
        }
        ifile.close();
        std::system(("rm " + solverResult).c_str());
        
        Assert<CE::AbortedMission>(isOK, "Aborted Solver");
    } catch(...) {
        throw;
    }
    return costOfSoln >= 0;
}

//NATHAN

void print_SP_plan(const Input& input, vector<int> &varAssignments)
{
    //this must read the variable info file... 
    string varFileName = input.VarFile();
    
    ifstream ifile(varFileName.c_str());
    if ( !ifile ) {
        cerr << "Error: Cannot open variable assignment file...\n";
        exit(1);
    }
    string tStr;
    int vNo;
    string vType;
    int tStep;
    int tCost;
    string aName;
    int nVars = -1;
    
    //Read the variable file produced by bb along with the cnf
    ifile >> nVars;
    assert(nVars != -1);
    
    vector<string> typeArray(nVars + 1);
    vector<int>    costArray(nVars + 1);
    vector<string> nameArray(nVars + 1);
    vector<int>    timeArray(nVars + 1);
    while (ifile >> tStr) {
        if (tStr != "v") cout << "tSTR: " << tStr << endl;
        assert(tStr == "v");
        ifile >> vNo;
        ifile >> vType; assert(vType == "p" || vType == "a" || vType == "n");
        ifile >> tStep;
        ifile >> tCost;
        aName = "";
        while (true) {
            ifile >> tStr;
            if (tStr == "-1") break;
            aName += tStr + " ";
        }
        typeArray[vNo] = vType;
        costArray[vNo] = tCost;
        nameArray[vNo] = aName;
        timeArray[vNo] = tStep;
        if (ifile.eof()) break;
    }

    //Make another cnf file with the variables appended with string 
    //representations of their names
    #if 0 
        
        ifstream cfile(input.CNF().c_str());
        if ( !cfile ) {
            cerr << "Error: Cannot open cnf file...\n";
            exit(1);
        }
        
        ofstream cofile((input.CNF() + "_").c_str());
        if ( !cofile ) {
            cerr << "Error: Cannot open cnfo file...\n";
            exit(1);
        }
        
        string ttS;
        while (true) {
            cfile >> ttS;
            if (ttS == "cnf") {
                cfile >> ttS;
                cfile >> ttS;
                break;
            }
        }
        //read the clauses
        int ttInt;
        while (cfile >> ttInt) {
            if (ttInt == 0) {
                cofile << "0\n";
            } else {
                cofile << ttInt << "(" << nameArray[abs(ttInt)] << ") ";
            }
            
        }
        cofile.close();
    #endif

    //Print the plan
    cout << "SAT Plan:" << endl;
    for (unsigned int aPos = 0; aPos < varAssignments.size(); aPos++) {
        int tV = varAssignments[aPos];
        if (tV < 0 || typeArray[tV] != "a") continue;
        cout << timeArray[tV] << ": (" << tV << ") " << nameArray[tV] << 
            " [" << costArray[tV] << "]\n";

    }
}

//Modified by NATHAN
//========
// main()
//========
int main(int argc, char* argv[])
{
    try {
        int bestScore = -1;
        // Some locals
        Timer timer;
        timer.Start();        
        Input input(argc, argv);
        
        long maxLevel = input.MaxLevel();
        long inputLevel = input.Level();
        bool firstOnly = true;
        bool result = false;
        if ( inputLevel < 0 ) {
            inputLevel = 1;
            firstOnly = false;
        } else if ( maxLevel < inputLevel ) {
            maxLevel = inputLevel;
        }
        
        cout << "CO-PLAN: Nathan Robinson, Charles Gretton, Duc-Nghia Pham, May 2008\n";
        cout << "Domain: " << input.Domain() << endl;
        cout << "Problem: " << input.Problem() << endl;
    
        vector<int> varAssignments;
        
        int prevBestScore = -1;
        int expandFailures = 0;
        int bestLevel = -1;
        for ( ; inputLevel <= maxLevel; ++inputLevel ) {
        
            cout << "Level: " << inputLevel << endl;
            varAssignments.clear();
         
            if ( firstOnly ) { // looking for specific plan-level only
                if ( inputLevel != input.Level() ) {
                    cout << "CNF Unsat." << endl;
                    break;
                }
            }
            
            // Build CNF encoding
            Status status = GenEncoding(input, inputLevel, prevBestScore);
            if ( status == UNSAT ) {// Couldn't reach goal state
                cout << "Goal not reached." << endl;
                continue;  
            }          
            Assert<CE::ErrorOccurred>(status == SAT, "Building CNF Failed.\n");
            // Throw solver at real CNF sentence
            bool solved = false;
            try {
                cout << "Solving CNF..." << endl;
                solved = SolveCNF(input, bestScore, varAssignments);
            } catch(CE::TimeOut&) {
                cout << "Timeout or memout: " << " at level " << inputLevel << endl;
            }
            if ( !solved ) {
                //NATHAN
                //This code can be reintroduced if coplan should contntinue past the
                //step optimal horizon until a better solution cannot be found and the
                //number of failures is exausted
                //The lines below the symbol *11 in cnfout.cc and below in this file 
                //should also be uncommeted
                //if (prevBestScore > -1) {
                 //   cout << "No better solutions at this horizon.\n";
                //    result = solved;
                //    bestScore = prevBestScore;
                //} else {
                    cout << "CNF Unsat." << endl;
                    continue; // CNF Unsat: try another layer
                //*11
                //}
            } else {
                result = solved;
                // *11
                //if (varAssignments.size() > 0)  {
                print_SP_plan(input, varAssignments);
                // *11
                //} else {
                //    cout << "No better solutions at this horizon.\n";
                //    bestScore = prevBestScore;
                //}
            }
            // **1
            //#define FAILURE_LIMIT 1
            //if (prevBestScore == -1 || bestScore < prevBestScore) {
                prevBestScore = bestScore;
                bestLevel = inputLevel;
                // **1
                //continue;
            //} else if (expandFailures++ < FAILURE_LIMIT) continue;
            break;
        }
        timer.Stop();
        
        // Display some stats
        cout << "\nCNF Sat. Solved in " << bestLevel << " layers, Score: " << prevBestScore << endl;
        cout << " Time:"; cout << timer << std::endl;
        

        //Solve with Charles' planner
        local_parse_function(input.Domain(), input.Problem());
        

        if (prevBestScore < 0) {
            cout << "Error: SatPlan score is undefined... using 1000 by default.\n";
            prevBestScore = 1000;
        }
        Planning::CostPlanner planner(problem, prevBestScore+1);
        
        cout << "Exausting the state space...\n";
        timer.Reset();
        timer.Start();
        planner();
        timer.Stop();
        cout << "Time:" << timer << endl;
        
        //Attempt to open the solution file
        ofstream solnfile(input.Solution().c_str());
        if ( !solnfile ) {
            cerr << "Can not open solution file: " << input.Solution() << endl;
            exit(1);
        }
        
        const vector<Planning::GroundActionWithCost>& plan = planner.getPlan();
        int aCtr = 0;
        int tCost = 0;
        cout << "Plan:\n";
        for (vector<Planning::GroundActionWithCost>::const_reverse_iterator pIt = 
            plan.rbegin(); pIt != plan.rend(); pIt++) {
            cout << aCtr << ": " << pIt->name << " [" << pIt->cost << "]" << endl;    
            solnfile << pIt->name << endl;
            ++aCtr;
            tCost += pIt->cost;
        }
        cout << "Total Cost: " << tCost << endl;

        return(GOOD);
    } catch(CE::Usage& u) {
        std::cerr << std::endl << u.GetMessage() << std::endl;
        char* tmp[2];
        Input cheat(0, tmp);
        std::cerr << cheat << std::endl;
        return(BADUSAGE);
    } catch(Input::MakeList&) { // thrown by Input upon user-request
        return(GOOD);
    } catch(Unsolved& uns) {
        std::cout << std::endl << uns.GetMessage() << std::endl;
        return(NOSOLVE);
    } catch(CE::MemoryLimit& ml) {
        std::cerr << std::endl << ml.GetMessage() << std::endl;
        return(NOMEMORYTIME);
    } catch(CE::TimeOut& tm) {
        std::cerr << std::endl << tm.GetMessage() << std::endl;
        return(NOMEMORYTIME);
    } catch(Exc::BaseException& eb) {
        std::cerr << std::endl << eb.GetMessage() << std::endl;
    } catch(...) {
        std::cerr << std::endl << "Unknown runtime error..." << std::endl;
    }
    return(NOGOOD);
}

namespace {

struct absComp : public std::binary_function<bool, long, long> {
    bool operator()(long a, long b) {
        return(abs(a) < abs(b));
    }
};

// CheckProgramTime()
void CheckProgramTime(const Input& in, const Timer& t) {
    Assert<CE::TimeOut>(t.UserTime() <= in.TimeAlloted(), "Program Timeout");
}

} // unnamed namespace

