/*============================================================================//
    Input.cpp: Implementation of Input.h.

    Implementation: Shane J. Neph, June 2004, University of Washington
//============================================================================*/

//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.

// Files included
#include "ByLine.h"
#include "Input.h"

// unix-specific Files included
#include <sys/resource.h>

//==============
// checkInput()
//==============
void Input::checkInput(int argc, char* argv[]) {
    // Assert some things; stuff applicable values in params_
    using StringAlgs::Trim;
    using CE::Usage;
    typedef std::map<std::string, std::string>::iterator Iter;

    for ( long idx = 1; idx < argc; ) {
        // Option
        std::string nextA = Trim(std::string(argv[idx]));
        bool dummy = (nextA[0] == '-');
        Assert<Usage>(dummy, "Expected -option, received: " + nextA);
        Iter iter = params_.find(nextA);
        Assert<Usage>(iter != params_.end(), "Unknown option: " + nextA);
        Assert<Usage>(iter->second.empty(), "more than 1 def'n: " + nextA);
        
        // Value
        Assert<Usage>(++idx < argc, "No value found for: " + nextA);
        std::string nextB = Trim(std::string(argv[idx]));
        Assert<Usage>(!nextB.empty(), "No value found for: " + nextA);
        dummy = (nextB[0] != '-');
        Assert<Usage>(dummy, "No value found for option: " + nextA);

        iter->second = nextB;
        ++idx;
    }
    makeConversions();
}

//===================
// makeConversions()
//===================
//Modidifed by NATHAN - some arguments removed some added (as indicated).
void Input::makeConversions() {
    typedef std::map<std::string, std::string> MapType;
    typedef MapType::iterator Iter;

    using CE::Usage;
    using StringAlgs::IsInteger;

    // -solver
    bool dummy;
    Iter iter;
    iter = params_.find("-solver");

    if ( iter->second.empty() ) { // Select default solver
        satSolver_ = DEFAULT_EXE_NAME;
    }
    else { // Use specified solver (ensure it exists)
        satSolver_ = iter->second;
    }
    
    //ensure that the solver exists...
    bool solverOK = true;
    std::ifstream test(satSolver_.c_str());
    if ( !test ) solverOK = false;
    std::vector<std::string> toCheck;
    std::copy(std::istream_iterator<ByLine>(test),
                std::istream_iterator<ByLine>(),
                std::back_inserter(toCheck));
    if (toCheck.empty()) solverOK = false;
    Assert<NoSolvers>(solverOK, "No binary or script found for: " + satSolver_);

    iter = params_.find("-solverlim");
    if (!iter->second.empty() ) {
        Assert<Usage>(dummy = IsInteger(iter->second), 
                             "Expected +Int value for -solverlim");
        solverLim_ = convert<long>(iter->second);
        Assert<Usage>(solverLim_ > 0, "Expect +Int value for -solverLim");
    } else solverLim_ = 120;

    // -globaltime 
    iter = params_.find("-globaltime");
    if ( iter->second.empty() ) {
        time_ = 30*60; // min*s/min
        iter->second = convert<std::string>(time_);
    }
    else { // receive from user in minutes
        Assert<Usage>(dummy = IsInteger(iter->second), 
                             "Expected Int value for -globaltime");
        time_ = convert<long>(iter->second)*60; // min*s/min
        Assert<Usage>(time_ > 0, "Expect +Int value for -globaltime");
    }

    // -globalmemory
    iter = params_.find("-globalmemory");
    if ( iter->second.empty() ) {
        mem_ = (long)(10 * 1024 * 1024 * 1024); 
        iter->second = convert<std::string>(mem_);
    } else {
        Assert<Usage>(dummy = IsInteger(iter->second), 
                             "Expected +Int value for -globalmemory");
        mem_ = convert<long>(iter->second);
        Assert<Usage>(mem_ > 0, "Expected +Int value for -globalmemory");
    }

    // -timeout
    iter = params_.find("-solverlim");
    if ( iter->second.empty() ) {
        solverLim_ = 1800;
        iter->second = convert<std::string>(solverLim_);
    }
    else { // receive from user in seconds
        Assert<Usage>(dummy = IsInteger(iter->second), 
                             "Expected Int value for -timeout");
        solverLim_ = convert<long>(iter->second);
        Assert<Usage>(solverLim_ > 0, "Expected +Int value for -timeout");
    }

    // -level
    iter = params_.find("-level");
    if ( iter->second.empty() )
        level_ = -1;    
    else {
        Assert<Usage>(dummy = IsInteger(iter->second),
                             "Expected Int value for -level");
        level_ = convert<long>(iter->second);
        Assert<Usage>(level_ > 0, "Expected +Int value for -level");
    }

    // -maxlevel
    iter = params_.find("-maxlevel");
    if ( iter->second.empty() ) {
        maxLevel_ = 1000; // arbitrarily large
        iter->second = convert<std::string>(maxLevel_);
    }
    else {
        Assert<Usage>(dummy = IsInteger(iter->second),
                             "Expected Int value for -maxlevel");
        maxLevel_ = convert<long>(iter->second);
        Assert<Usage>(maxLevel_ > 0, "Expected +Int value for -maxlevel");
    }
    Assert<Usage>(maxLevel_ >= level_, 
                  "Expect -maxlevel's value >= -level's value");

    // -cnf
    iter = params_.find("-cnf");
    cnf_ = iter->second.empty() ? DEFAULT_CNF_NAME : iter->second;
    iter->second = cnf_;


    // -var
    iter = params_.find("-var");
    varFile_ = iter->second.empty() ? DEFAULT_VAR_FILE_NAME : iter->second;
    iter->second = varFile_;

    // -domain
    iter = params_.find("-domain");
    Assert<Usage>(!iter->second.empty(), "must define -domain");
    domain_ = iter->second;

    // -problem
    iter = params_.find("-problem");
    Assert<Usage>(!iter->second.empty(), "must define -problem");
    problem_ = iter->second;

    // -solution (must follow -problem)
    iter = params_.find("-solution");
    solution_ = iter->second.empty() ? DEFAULT_SOLUTION_NAME : iter->second;
    iter->second = solution_;

}

//=============
// setMemory()
//=============
void Input::setMemory() {
    rlimit limit;
    limit.rlim_cur = mem_;
    limit.rlim_max = mem_;
    setrlimit(RLIMIT_AS, &limit);     
}

//================
// unknownUsage()
//================
std::string Input::unknownUsage(const std::string& str) {
    return("Unknown usage... " + str);
}
