/*============================================================================//
    Input.h: header file for Input UDT

    Purpose is to process the user inputs and provide defaults when certain
     inputs are left undefined.  Static member function Revision() maintains
     the revision of the entire SatPlan program.

    NOTE: Use to 'setrlimit' for global time here, but it seemed to interfere 
          somewhat with ulimit settings in other parts of the program.  Now,
          SatPlan checks TimeAllotted() against user time.  This method works
          because we do not fork() other processes any longer (we use system()
          calls).  However, it is now possible for our program to run up to
          (-globaltime + -timeout) instead of throwing at -globaltime.  The
          good news is that we will still catch the problem ( > -globaltime )
          and report a program timeout...bad news is that we may run a little
          longer than we wanted to.  Some constraints on the ratio of -timeout
          as compared to -globaltime have been added to reduce this problem.

    Implementation: Shane J. Neph, June 2004, University of Washington
//============================================================================*/

//Modified by Nathan Robinson, May 2008, Griffith University
// Copyright of the modifications (c) 2008 National ICT Australia Limited (NICTA)
//
// Author: Nathan Robinson (nathan.robinson@nicta.com.au)
//
// Redistribution and use in source and binary forms, with or without
// modification and only for non-commercial research and educational
// purposes are permitted provided that the conditions in the
// accompanying file "licence.txt" are met.


// Macro Guard
#ifndef INPUT_CSE473_H
#define INPUT_CSE473_H

// Files included
#include "Assertion.h"
#include "CommonErrors.h"
#include "Conversion.h"
#include "StringAlgorithms.h"

//NATHAN ROBINSON
#define DEFAULT_SAT_RESULT_NAME "satsolver.res"
#define DEFAULT_EXE_NAME "rsat_opt_cost"
#define DEFAULT_SOLUTION_NAME "out.solution"
#define DEFAULT_CNF_NAME "out.cnf"
#define DEFAULT_VAR_FILE_NAME "out.vars"

//=======
// Input
//=======
struct Input {
    typedef Exc::Exception<1001> MakeList;
    typedef Exc::Exception<1002> NoSolvers;

    // Constructor
    //Modified by NATHAN
    Input(int argc, char* argv[]) : domain_(""), problem_("") {
        std::string empty("");
        params_.insert(std::make_pair(std::string("-cnf"), empty));
        params_.insert(std::make_pair(std::string("-domain"), empty));
        params_.insert(std::make_pair(std::string("-globalmemory"), empty));
        params_.insert(std::make_pair(std::string("-globaltime"), empty));
        params_.insert(std::make_pair(std::string("-level"), empty));
        params_.insert(std::make_pair(std::string("-maxlevel"), empty));
        params_.insert(std::make_pair(std::string("-problem"), empty));
        params_.insert(std::make_pair(std::string("-solution"), empty));
        params_.insert(std::make_pair(std::string("-solver"), empty));
        params_.insert(std::make_pair(std::string("-solverlim"), empty));
        params_.insert(std::make_pair(std::string("-var"), empty));
        if ( 0 == argc ) // dummy Input object constructed
            return;
        
        checkInput(argc, argv);
        setMemory();
    }

    // Public Member Functions
    std::string AllInputs(const std::string& add = "") const {
        std::string toRtn;
        std::map<std::string, std::string>::const_iterator i = params_.begin();
        while ( i != params_.end() ) {
            if ( !i->second.empty() )
                toRtn += (add + " " + i->first + " " + i->second + "\n");
            ++i;
        }
        return(toRtn);
    }

    std::string CNF() const {
        return(cnf_);
    }

    std::string Domain() const {
        return(domain_);
    }
    
    long Level() const {
        return(level_);
    }

    long MaxLevel() const {
        return(maxLevel_); 
    }

    std::string Problem() const {
        return(problem_);
    }
    //NATHAN
    std::string Program() const {
        return(satSolver_);
    }

    long Rand() const {
        return(rand());
    }
    
    std::string Solution() const {
        return(solution_);
    }
    long TimeAlloted() const {
        return(time_);
    }
    std::string VarFile() const {
        return(varFile_); // bb produces this file
    }
    //NATHAN
    int SolverLim() const {
        return(solverLim_);
    }
    //Modified by NATHAN
    friend std::ostream& operator<<(std::ostream& os, const Input&) {
        os << "==================================================="<< std::endl;
        os << "Usage: [...] is optional"                           << std::endl;
        os << "==================================================="<< std::endl;
        os << "-problem X (:string - problem filename)"            << std::endl;
        os << "-domain X (:string - domain filename)"              << std::endl;
        os << "[-cnf X (:string - CNF encoding output filename)]"  << std::endl;
        os << "[-globalmemory X (:integer - bytes)]"               << std::endl;
        os << "[-globaltime X (:integer - minutes)]"               << std::endl;
        os << "[-level X (:integer - try this plan level only)]"   << std::endl;
        os << "[-maxlevel X (:integer - try <= this plan level)]"  << std::endl;
        os << "[-solution X (:string - solution filename)]"        << std::endl;
        os << "[-solver X (:string - sat solver name)]"            << std::endl;
        os << "[-timeout X (:integer - seconds before new rand())]"<< std::endl;
        os << "[-varfile X (:string - variable file name)]"        << std::endl;
        os << "[-solverlim X (:integer - tlimit per cnf, (d:120))]"<< std::endl;
        return(os);
    }

private:
    void checkInput(int argc, char* argv[]);
    void makeConversions();
    void setMemory();
    std::string unknownUsage(const std::string& str);
private:
    //NATHAN
    long time_, mem_, level_, maxLevel_, solverLim_;
    std::string cnf_, domain_, problem_, solution_, satSolver_, varFile_;
    std::map<std::string, std::string> params_;
};

#endif // INPUT_CSE473_H

