#######################################################################
#
# Author: Malte Helmert (helmert@informatik.uni-freiburg.de)
# (C) Copyright 2003-2004 Malte Helmert
# Modified by: Gabi Roeger and Silvia Richter (silvia.richter@nicta.com.au)
# (C) Copyright 2008: Gabi Roeger and NICTA
#
# This file is part of LAMA.
#
# LAMA is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the license, or (at your option) any later version.
#
# LAMA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import conditions
import pddl_types
import f_expression

def cartesian_product(*sequences):
  # TODO: Also exists in tools.py outside the pddl package (defined slightly
  #       differently). Not good. Need proper import paths.
  if not sequences:
    yield ()
  else:
    for tup in cartesian_product(*sequences[1:]):
      for item in sequences[0]:
        yield (item,) + tup

def parse_effects(alist, result):
  """Parse a PDDL effect (any combination of simple, conjunctive, conditional, and universal)."""
  tmp_effect = parse_effect(alist)
  #print "original effect"
  #tmp_effect.dump() 
  normalized = tmp_effect.normalize()
  cost_eff, rest_effect = normalized.extract_cost()
  #print "cost of action is", cost_expr
  #print "normalized effect"
  #normalized.dump()
  add_effect(rest_effect, result)
  if cost_eff:
    return cost_eff.effect, tmp_effect
  else:
    return None, tmp_effect
 
def parse_effect(alist):
    tag = alist[0]
    if tag == "and":
      return ConjunctiveEffect([parse_effect(eff) for eff in alist[1:]])
    elif tag == "forall":
      assert len(alist) == 3
      parameters = pddl_types.parse_typed_list(alist[1])
      effect = parse_effect(alist[2])
      return UniversalEffect(parameters, effect)
    elif tag == "when":
      assert len(alist) == 3
      condition = conditions.parse_condition(alist[1])
      effect = parse_effect(alist[2])
      return ConditionalEffect(condition, effect)
    elif tag == "increase":
      assert len(alist) == 3
      assert alist[1] == ['total-cost']
      assignment = f_expression.parse_assignment(alist)
      return CostEffect(assignment)
    else:
        return SimpleEffect(conditions.parse_literal(alist))
      

def add_effect(tmp_effect, result):
  """tmp_effect has the following structure:
     [ConjunctiveEffect] [UniversalEffect] [ConditionalEffect] SimpleEffect."""
  
  if isinstance(tmp_effect, ConjunctiveEffect):
    for effect in tmp_effect.effects:
      add_effect(effect, result)
    return
  else:
    parameters = []
    condition = conditions.Truth()
    if isinstance(tmp_effect, UniversalEffect):
      parameters = tmp_effect.parameters
      if isinstance (tmp_effect.effect, ConditionalEffect):
        condition = tmp_effect.effect.condition
        assert isinstance(tmp_effect.effect.effect, SimpleEffect)
        effect = tmp_effect.effect.effect.effect
      else:
        assert isinstance(tmp_effect.effect, SimpleEffect)
        effect = tmp_effect.effect.effect
    elif isinstance(tmp_effect, ConditionalEffect):
      condition = tmp_effect.condition
      assert isinstance(tmp_effect.effect, SimpleEffect)
      effect = tmp_effect.effect.effect
    else:
      assert isinstance(tmp_effect, SimpleEffect)
      effect = tmp_effect.effect
    assert isinstance(effect, conditions.Literal)
    # Check for contradictory effects
    new_effect = Effect(parameters, condition, effect)
    contradiction = Effect(parameters, condition, effect.negate())
    if not contradiction in result:
      result.append(new_effect)
    else:
      # We use add-after-delete semantics, keep positive effect
      if isinstance(contradiction.literal, conditions.NegatedAtom):
        result.remove(contradiction)
        result.append(new_effect)

class Effect(object):
  def __init__(self, parameters, condition, literal):
    self.parameters = parameters
    self.condition = condition
    self.literal = literal
  def __eq__(self, other):
    return (self.__class__ is other.__class__ and
            self.parameters == other.parameters and
            self.condition == other.condition and
            self.literal == other.literal)
  def dump(self):
    indent = "  "
    if self.parameters:
      print "%sforall %s" % (indent, ", ".join(map(str, self.parameters)))
      indent += "  "
    if self.condition != conditions.Truth():
      print "%sif" % indent
      self.condition.dump(indent + "  ")
      print "%sthen" % indent
      indent += "  "
    print "%s%s" % (indent, self.literal)
  def uniquify_variables(self, type_map):
    renamings = {}
    self.parameters = [par.uniquify_name(type_map, renamings)
                       for par in self.parameters]
    self.condition = self.condition.uniquify_variables(type_map, renamings)
    self.literal = self.literal.rename_variables(renamings)
  def constants(self):
    result = set()
    result |= self.condition.constants()
    result |= self.literal.constants()
    return result
  def unground(self, mapping):
    return self.literal._unground(mapping)
#  def build_prolog_rules(self, prog, rule_body):
#    assert not self.literal.negated, "Something fishy"
#    assert self.condition == conditions.Truth(), "Not implemented"
#    prog.add_rule(rule_body, self.literal)
#  def to_untyped_strips(self):
#    # Used by pddl_to_prolog.py for generating Horn rules.
#    # TODO: This is somewhat hackish and incomplete for ADL.
#    #       Should be replaced by something more reasonable.
#    assert not self.literal.negated, "Something strange here"
#    conditions = [par.to_untyped_strips() for par in self.parameters]
#    conditions += self.condition.to_untyped_strips()
#    return conditions, self.literal
  def instantiate(self, var_mapping, init_facts, fluent_facts,
                  objects_by_type, result):
    if self.parameters:
      var_mapping = var_mapping.copy() # Will modify this.
      object_lists = [objects_by_type.get(par.type, [])
                      for par in self.parameters]
      for object_tuple in cartesian_product(*object_lists):
        for (par, obj) in zip(self.parameters, object_tuple):
          var_mapping[par.name] = obj
        self._instantiate(var_mapping, init_facts, fluent_facts, result)
    else:
      self._instantiate(var_mapping, init_facts, fluent_facts, result)
  def _instantiate(self, var_mapping, init_facts, fluent_facts, result):
    condition = []
    try:
      self.condition.instantiate(var_mapping, init_facts, fluent_facts, condition)
    except conditions.Impossible:
      return
    effects = []
    self.literal.instantiate(var_mapping, init_facts, fluent_facts, effects)
    assert len(effects) <= 1
    if effects:
      result.append((condition, effects[0]))
  def relaxed(self):
    if self.literal.negated:
      return None
    else:
      return Effect(self.parameters, self.condition.relaxed(), self.literal)
  def simplified(self):
    return Effect(self.parameters, self.condition.simplified(), self.literal)


        #def to_pddl(self, indent = "  "):
        #pddl_string = "%s(%s\n" % (indent, self._to_pddl(indent = ""))
        #for part in self.parts:
        #    pddl_string += part.to_pddl(indent + "     ") + "\n"
        #return pddl_string + indent + ")"
      
class ConditionalEffect(object):
  def __init__(self, condition, effect):
    if isinstance(effect, ConditionalEffect):
      self.condition = conditions.Conjunction([condition, effect.condition])
      self.effect = effect.effect
    else:
      self.condition = condition
      self.effect = effect
  def dump(self, indent="  "):
    print "%sif" % (indent)
    self.condition.dump(indent + "  ")
    print "%sthen" % (indent)
    self.effect.dump(indent + "  ")
  def constants(self):
    result = set()
    result |= self.condition.constants()
    result |= self.effect.constants()
    return result
  def unground(self, mapping):
    return ConditionalEffect(self.condition.unground(mapping),
                             self.effect.unground(mapping))
  def normalize(self):
    norm_effect = self.effect.normalize()
    if isinstance(norm_effect, ConjunctiveEffect):
      new_effects = []
      for effect in norm_effect.effects:
        assert isinstance(effect, SimpleEffect) or isinstance(effect, ConditionalEffect)
        new_effects.append(ConditionalEffect(self.condition, effect))
      return ConjunctiveEffect(new_effects)
    elif isinstance(norm_effect, UniversalEffect):
      child = norm_effect.effect
      cond_effect = ConditionalEffect(self.condition, child)
      return UniversalEffect(norm_effect.parameters, cond_effect)
    else:
      return ConditionalEffect(self.condition, norm_effect)
  def extract_cost(self):
    return None, self

class UniversalEffect(object):
  def __init__(self, parameters, effect):
    if isinstance(effect, UniversalEffect):
      self.parameters = parameters + effect.parameters
      self.effect = effect.effect
    else:
      self.parameters = parameters
      self.effect = effect
  def dump(self, indent="  "):
    print "%sforall %s" % (indent, ", ".join(map(str, self.parameters)))
    self.effect.dump(indent + "  ")
  def constants(self):
    return self.effect.constants()
  def unground(self, mapping):
    return UniversalEffect(self.parameters,
                           self.effect.unground(mapping))
  def normalize(self):
    norm_effect = self.effect.normalize()
    if isinstance(norm_effect, ConjunctiveEffect):
      new_effects = []
      for effect in norm_effect.effects:
        assert isinstance(effect, SimpleEffect) or isinstance(effect, ConditionalEffect)\
               or isinstance(effect, UniversalEffect)
        new_effects.append(UniversalEffect(self.parameters, effect))
      return ConjunctiveEffect(new_effects)
    else:
      return UniversalEffect(self.parameters, norm_effect) 
  def extract_cost(self):
    return None, self
    
class ConjunctiveEffect(object):
  def __init__(self, effects):
    flattened_effects = []
    for effect in effects:
      if isinstance(effect, ConjunctiveEffect):
        flattened_effects += effect.effects
      else:
        flattened_effects.append(effect)
    self.effects = flattened_effects
  def dump(self, indent="  "):
    print "%sand" % (indent)
    for eff in self.effects:
      eff.dump(indent + "  ")
  def to_pddl(self, indent = "  "):
    pddl_string = "%s(and\n" % (indent)
    for effect in self.effects:
      pddl_string += effect.to_pddl(indent + "     ") + "\n"
    return pddl_string + indent + ")"
  def constants(self):
    result = set()
    for effect in self.effects:
      result |= effect.constants()
    return result
  def unground(self, mapping):
    return ConjunctiveEffect([effect.unground(mapping) for effect in self.effects])
  def normalize(self):
    new_effects = []
    for effect in self.effects:
      new_effects.append(effect.normalize())
    return ConjunctiveEffect(new_effects)
  def extract_cost(self):
    new_effects = []
    cost_effect = None
    for effect in self.effects:
      if isinstance(effect, CostEffect):
        cost_effect = effect
      else:
        new_effects.append(effect)
    return cost_effect, ConjunctiveEffect(new_effects)
                         
class SimpleEffect(object):
  def __init__(self, effect):
    self.effect = effect
  def dump(self, indent="  "):
    print "%s%s" % (indent, self.effect)
  def to_pddl(self, indent = "  "):
    return "%s%s" % (indent, self.effect.to_pddl())
  def constants(self):
    return self.effect.constants()
  def unground(self, mapping):
    return SimpleEffect(self.effect._unground(mapping))
  def normalize(self):
    return self
  def extract_cost(self):
    return None, self
  
class CostEffect(object):
  def __init__(self, effect):
    self.effect = effect
  def dump(self, indent="  "):
    print "%s%s" % (indent, self.effect)
  def to_pddl(self, indent = "  "):
    return "%s%s" % (indent, self.effect.to_pddl())
  def constants(self):
    return set()
  def unground(self, mapping):
    return self
  def normalize(self):
    return self
  def extract_cost(self):
    return self, None # this would only happen if
  #an action has no effect apart from the cost effect
